#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb  5 09:19:46 2019

@author: stan
"""
from itertools import product

class RedBooleana:
    
    def __init__(self,vertices,aristas,reglas,tablas=[]):
        self.vertices = vertices
        self.n = len(vertices)
        self.aristas = aristas
        self.reglas = reglas
        self.tablas = tablas
        self.st_table = None
        self.estado = tuple(False for i in range(self.n))
        self.pos = {vertices[i]:i for i in range(len(vertices))}
        
    
    def set_estado(self,x):
        self.estado = x
        
    def update(self,table,estado):
        reg_vals = tuple(estado[self.pos[v]] for v in table.variables)
        return table[reg_vals]
        
    def next_state(self,estado = None):
        if not estado:
            estado = self.estado
        d = {v:s for v,s in zip(self.vertices,estado)}
        if len(self.tablas)==0:
            self.estado = tuple(r(d) for r in self.reglas)
        else:
            self.estado = \
            tuple(self.update(t,estado)\
                  for g,t in sorted(self.tablas.items(),\
                                    key =lambda t:self.pos[t[0]]))
        return self.estado
        
    @staticmethod
    def state_as_string(s):
        return "".join(map(lambda x:"1" if x else "0",s))
        
    def iterar(self,n):
        for i in range(n):
            print(RedBooleana.state_as_string(self.estado))
            self.next_state()
            
    def compute_state_transition_table(self):
        t = [False,True]
        g = (t for i in range(len(self.vertices))) 
        i = list(product(*g))
        tabla = {izq:der for izq,der in zip(i,(self.next_state(s) for s in i))}
        self.st_table = tabla
        
    def compute_fixed_point_attractors(self):
        if not self.st_table:
            self.compute_state_transition_table()
        return [i for i,d in self.st_table.items() if i==d]
    
    
    def get_attrs(self,trayectoria,rep_s):
        trayectoria.reverse()
        for indice,estado in enumerate(trayectoria):
            if estado==rep_s:
                attr = trayectoria[:indice+1]
        attr.reverse()
        trayectoria.reverse()
        return tuple(attr)
        
    
    
    def compute_attractors(self):
        if not self.st_table:
            self.compute_state_transition_table()
        cuencas = {}
        visitados = set()
        profundidad = {}
        prede = {}
        for i,d in self.st_table.items():
            if d in prede:
                prede[d].append(i)
            else:
                prede[d] = [i]
            trayectoria = []
            if i in visitados:
                continue
            visitados.add(i)
            trayectoria.append(i)
            while True:
                if d in visitados:
                    if d in trayectoria:
                        attractor = self.get_attrs(trayectoria,d)
                        cuencas.update({s:attractor for s in trayectoria}) 
                        idx = trayectoria.index(d)
                        profundidad.update(
                                {trayectoria[i]:(idx-i if i<=idx else 0)\
                                for i in range(len(trayectoria))})
                        break
                    else:
                        attractor = cuencas[d]
                        cuencas.update({s:attractor for s in trayectoria})
                        idx = profundidad[d]+len(trayectoria)
                        profundidad.update(
                                {trayectoria[i]:(idx-i if i<=idx else 0)\
                                for i in range(len(trayectoria))})
                        break
                trayectoria.append(d)
                i = d
                visitados.add(i)
                d = self.st_table[i]
                
        resultado = {}
        for estado,attr in cuencas.items():
            if attr in resultado:
                resultado[attr] += [estado]
            else:
                resultado[attr] = [estado]
                
        return (resultado,profundidad,prede)
                
        
            
    def get_truth_table(self):
        s = ""
        sep = " "
        if not self.st_table:
            self.compute_state_transition_table()
        
        encabezado = sep.join(self.vertices)
        encabezado += sep+"|"+sep
        encabezado += sep.join(self.vertices)
        encabezado += "\n"
        encabezado += "-"*len(encabezado)
        encabezado += "\n"
        s += encabezado
        for tupla in sorted(self.st_table.items()):
            l = ""
            izq,der = tupla
            l += sep.join("1" if i else "0" for i in izq)
            l += sep+"|"+sep
            l += sep.join("1" if i else "0" for i in der)
            l += "\n"
            s += l
        return s
    
    @staticmethod
    def find_regulations(tabla):
        regulaciones= []
        for p,v in enumerate(tabla.variables):
            positiva = False
            negativa = False
            orden = sorted(tabla.table.items(),
            key=lambda t:t[0][:p]+t[0][p+1:]+t[0][p:p+1])
            for i in range(2**(tabla.n-1)):
                f0 = tabla.table[orden[i*2][0]]
                f1 = tabla.table[orden[i*2+1][0]]
                # vemos si hay regulación positiva
                if f1-f0 == 1 and not positiva:
                    regulaciones.append((v,tabla.name,'+'))
                    positiva = True
                    if negativa:
                        break
                elif f1-f0 == -1 and not negativa:
                    regulaciones.append((v,tabla.name,'-'))
                    negativa = True
                    if positiva:
                        break
        return regulaciones
        
    def __str__(self):
        s = '( '
        for nombre,valor in zip(self.vertices,self.estado):
            s += nombre+'='+str(valor)+' '
        s += ')'
        return s
    
    def __repr__(self):
        return str(self)
        
        