#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  5 08:30:49 2019

@author: stan
"""
import matplotlib.pyplot as plt
import numpy as np
class VisualRB:
    
    def __init__(self,cuencas,r=1):
        """
        :param cuencas: diccionario con las cuencas
        la llave es el atractor y el valor la cuenca
        """
        self.cuencas = cuencas
        self.r = r
        self.pesos = {}
        
    def quitar_ciclos(self,predecesores,atractores):
        pred = predecesores.copy()
        for atractor in atractores:
            for estado in atractor:
                    pred[estado]=set(pred[estado])-set(atractor)
                    if pred[estado]==None:
                        del pred[estado]
        return pred
        
    def peso(self,estado,predecesores,atractores):
        """
        :param estado: estado inicial
        :param predecesores: lista de predecesores para cada estado
        es un diccionario donde la llave es el estado
        """
        pred = self.quitar_ciclos(predecesores,atractores)
        return self.peso_r(estado,pred,atractores)
    
    def peso_r(self,estado,predecesores,atractores):
        """
        :param estado: estado inicial
        :param predecesores: lista de predecesores para cada estado
        es un diccionario donde la llave es el estado
        """
        w = 0
        if estado in self.pesos:
            return self.pesos[estado]
        if estado in predecesores:
            for p in predecesores[estado]:
                w += 1 + self.peso_r(p,predecesores,atractores)
        self.pesos[estado] = w
        return w
        
        
    def coordenadas_r(self,
                      conjunto,
                      cuencas,
                      predecesores,
                      r=1,
                      offset=0,
                      angulo=2*np.pi,
                      coor={}):
        
        angulo = np.pi if angulo > np.pi else angulo
        suma = 0
        for estado in conjunto:
            if not estado in coor:
                suma += self.peso(estado,predecesores,cuencas.keys())+1


        for estado in conjunto:  
            if estado not in coor:
                peso = (self.peso(estado,predecesores,cuencas.keys())+1)
                delta =(angulo/suma)*peso
                x = r*np.cos(delta/2+offset)
                y = r*np.sin(delta/2+offset)
                coor[estado] = (x,y)
                if estado in predecesores:
                    coor = self.coordenadas_r(
                            predecesores[estado],
                            cuencas,
                            predecesores,
                            r+1,
                            offset,
                            delta,
                            coor)
                offset += delta
                
        return coor        
        
    def coordenadas(self,
                    conjunto,
                    cuencas,
                    predecesores,
                    r=1,
                    offset=0,
                    angulo=2*np.pi,
                    coor={}):
        
        atractores = list(cuencas.keys())
        pred = self.quitar_ciclos(predecesores,atractores)
        return self.coordenadas_r(
                conjunto,
                cuencas,
                pred,
                r,
                offset,
                angulo,
                coor)
            
    def show_coor(self,coor,atractor,basins,predecesores):
        n = len(basins[atractor])

        U = np.zeros((n,1))
        V = np.zeros((n,1))
        X = np.zeros((n,1))
        Y = np.zeros((n,1))
        idx = 0
        for estado in basins[atractor]:
            if estado in predecesores:
                for p in predecesores[estado]:
                    x_s,y_s = coor[estado]
                    x,y = coor[p]
                    U[idx] = x_s-x
                    V[idx] = y_s-y
                    X[idx] = x
                    Y[idx] = y
                    idx += 1
                
        plt.axis('equal')
        plt.quiver(
                X,
                Y,
                U,
                V,
                scale=1,
                scale_units='xy',
                width=0.005,
                facecolor='lightgray')
        plt.scatter(X,Y,color='blue')
        X,Y = zip(*map(lambda x:coor[x],atractor))
        plt.scatter(X,Y,color='red')
        
    